# OpenStack Cloud helper functions
import asyncio
import openstack


# Default values
DEFAULT_SUBNET_CIDR = "10.0.0.0/24"
DEFAULT_SUBNET_GATEWAY_IP = "10.0.0.1"
DEFAULT_ROLE_NAME = "member"
DEFAULT_EXTERNAL_NETWORK = "public1"
DEFAULT_ROUTER_NAME = "midgard-NAT"
DEFAULT_NETWORK_NAME = "midgard-net"
DEFAULT_SUBNET_NAME = "midgard-subnet"
DEFAULT_DNS_NAMESERVERS = ["1.1.1.1", "1.0.0.1"]
DEFAULT_IP_VERSION = 4
DEFAULT_KEYPAIR_NAME = "midgard-keypair"
DEFAULT_KEYPAIR_TYPE = "ssh"
DEFAULT_SG_DIRECTION = "ingress"
DEFAULT_SG_PROTOCOL = "tcp"
DEFAULT_SG_NAME = "default"
DEFAULT_SG_REMOTE_IP_PREFIX = "0.0.0.0/0"
DEFAULT_SERVER_NAME = "midgard-server"


def connect(
    auth_url: str = None,
    region_name: str = None,
    project_name: str = None,
    username: str = None,
    password: str = None,
    user_domain: str = None,
    project_domain: str = None,
):
    """Connect to an OpenStack client."""
    if not (auth_url or region_name or project_name or username or password):
        return openstack.connect(cloud="envvars")
    else:
        return openstack.connect(
            auth_url=auth_url,
            region_name=region_name,
            project_name=project_name,
            username=username,
            password=password,
            user_domain_name=user_domain,
            project_domain_name=project_domain,
        )


async def find_project(client: openstack.connection.Connection, project_name: str):
    """Find a project in Keystone database."""
    return await asyncio.to_thread(
        client.identity.find_project, project_name, ignore_missing=True
    )


async def find_user(client: openstack.connection.Connection, discord_user_id: str):
    """Find a user in Keystone database."""
    return await asyncio.to_thread(
        client.identity.find_user, discord_user_id, ignore_missing=True
    )


async def create_project(client: openstack.connection.Connection, project_name: str):
    """Create a new project in Keystone database."""
    return await asyncio.to_thread(client.identity.create_project, name=project_name)


async def create_user(
    client: openstack.connection.Connection, discord_user_id: str, **kwargs
):
    """Create a new user in Keystone database."""
    return await asyncio.to_thread(
        client.identity.create_user, name=discord_user_id, **kwargs
    )


async def update_user(
    client: openstack.connection.Connection,
    user: openstack.identity.v3.user.User,
    **kwargs,
):
    """Update a user in Keystone database."""
    return await asyncio.to_thread(client.identity.update_user, user, **kwargs)


async def set_default_roles(
    client: openstack.connection.Connection,
    user: openstack.identity.v3.user.User,
    project: openstack.identity.v3.project.Project,
) -> None:
    """Set default roles for a user in a project."""
    member_role = await asyncio.to_thread(client.identity.find_role, DEFAULT_ROLE_NAME)
    await asyncio.to_thread(
        client.identity.assign_project_role_to_user, project, user, member_role
    )


async def setup_default_network(
    client: openstack.connection.Connection,
    project: openstack.identity.v3.project.Project,
) -> None:
    """Setup default network for a project."""
    # Find default NAT project ID
    external_gateway = await asyncio.to_thread(
        client.network.find_network, DEFAULT_EXTERNAL_NETWORK
    )
    # Create default NAT router
    router = await asyncio.to_thread(
        client.network.create_router,
        name=DEFAULT_ROUTER_NAME,
        project_id=project.id,
        external_gateway_info={"network_id": external_gateway.id},
    )

    # Create network
    network = await asyncio.to_thread(
        client.network.create_network, project_id=project.id, name=DEFAULT_NETWORK_NAME
    )

    # Create subnet
    subnet = await asyncio.to_thread(
        client.network.create_subnet,
        network_id=network.id,
        cidr=DEFAULT_SUBNET_CIDR,
        project_id=project.id,
        name=DEFAULT_SUBNET_NAME,
        gateway_ip=DEFAULT_SUBNET_GATEWAY_IP,
        dns_nameservers=DEFAULT_DNS_NAMESERVERS,
        ip_version=DEFAULT_IP_VERSION,
    )

    # Add router interface
    await asyncio.to_thread(
        client.network.add_interface_to_router, router, subnet_id=subnet.id
    )


async def find_default_network(
    client: openstack.connection.Connection,
    **kwargs,
) -> openstack.network.v2.network.Network:
    """Find default network for a project."""

    return await asyncio.to_thread(
        client.network.find_network, DEFAULT_NETWORK_NAME, ignore_missing=True, **kwargs
    )


async def add_security_group_rule(
    client: openstack.connection.Connection,
    port: int,
    direction: str = DEFAULT_SG_DIRECTION,
) -> None:
    """Setup default security group for a project."""
    # Find security group
    security_group = await asyncio.to_thread(
        client.network.find_security_group,
        DEFAULT_SG_NAME,
    )

    # Create security group rule
    try:
        await asyncio.to_thread(
            client.network.create_security_group_rule,
            security_group_id=security_group.id,
            direction=direction,
            protocol=DEFAULT_SG_PROTOCOL,
            port_range_min=port,
            port_range_max=port,
            remote_ip_prefix=DEFAULT_SG_REMOTE_IP_PREFIX,
        )
    except openstack.exceptions.ConflictException as e:
        raise Exception(e.details.split(".")[0] + ".")


async def find_default_security_group(
    client: openstack.connection.Connection,
    **kwargs,
) -> openstack.network.v2.security_group.SecurityGroup:
    """Find default security group for a project."""
    return await asyncio.to_thread(
        client.network.find_security_group, DEFAULT_SG_NAME, **kwargs
    )


async def find_floating_ip(
    client: openstack.connection.Connection,
    **kwargs,
) -> openstack.network.v2.floating_ip.FloatingIP:
    """Retrun the first available floating IP"""
    return await asyncio.to_thread(client.available_floating_ip, **kwargs)


async def find_keypair(client: openstack.connection.Connection):
    """Find a keypair in a project."""
    return await asyncio.to_thread(
        client.compute.find_keypair,
        DEFAULT_KEYPAIR_NAME,
        ignore_missing=True,
    )


async def create_keypair(
    client: openstack.connection.Connection,
    public_key: str,
) -> None:
    """Create a new keypair for a project."""
    try:
        await asyncio.to_thread(
            client.compute.create_keypair,
            name=DEFAULT_KEYPAIR_NAME,
            public_key=public_key,
            type=DEFAULT_KEYPAIR_TYPE,
        )
    except openstack.exceptions.BadRequestException as e:
        raise Exception(e.details)


async def delete_keypair(
    client: openstack.connection.Connection,
    keypair: openstack.compute.v2.keypair.Keypair,
) -> None:
    """Delete a keypair from a project."""
    await asyncio.to_thread(client.compute.delete_keypair, keypair)


async def list_flavors(
    client: openstack.connection.Connection,
) -> list[openstack.compute.v2.flavor.Flavor]:
    """List all flavors."""
    flavors = await asyncio.to_thread(client.compute.flavors)
    return [flavor for flavor in flavors]


async def list_images(
    client: openstack.connection.Connection,
) -> list[openstack.image.v2.image.Image]:
    """List all images."""
    images = await asyncio.to_thread(client.image.images)
    return [image for image in images]


async def find_server(
    client: openstack.connection.Connection,
    name: str = None,
    **kwargs,
) -> openstack.compute.v2.server.Server:
    """Find a server in a project."""
    if name is None:
        name = DEFAULT_SERVER_NAME
    return await asyncio.to_thread(
        client.compute.find_server,
        name,
        ignore_missing=True,
        **kwargs,
    )


async def create_server(
    client: openstack.connection.Connection, **kwargs
) -> openstack.compute.v2.server.Server:
    """Create a new server in a project."""
    name = kwargs.pop("name", DEFAULT_SERVER_NAME)
    try:
        return await asyncio.to_thread(client.create_server, name=name, **kwargs)
    except openstack.exceptions.BadRequestException as e:
        raise Exception(e.details)



# Networking component for Midgard Discord Bot

import os
import aiohttp

CF_API_URL = "https://api.cloudflare.com/client/v4"
CF_ACCOUNT_ENDPOINT = "{CF_API_URL}/accounts/{CF_ACCOUNT_ID}"
CF_TUNNEL_ENDPOINT = "{CF_ACCOUNT_ENDPOINT}/cfd_tunnel/{CF_TUNNEL_ID}"
CF_DNS_ENDPOINT = "{CF_API_URL}/zones/{CF_ZONE_ID}/dns_records"


class Ingress:
    """A CloudFlare tunnel ingress rule."""

    def __init__(self, service: str, hostname: str = None, originRequest: dict = None):
        """Initialize the ingress rule."""
        self.service = service
        self.hostname = hostname
        self.originRequest = originRequest

    def __repr__(self) -> str:
        """Return the representation of the ingress rule."""
        if self.hostname:
            return f"<Ingress {self.hostname} --> {self.service} {self.originRequest}>"
        else:
            return f"<Ingress {self.service}>"


def get_client() -> aiohttp.ClientSession:
    """Get a CloudFlare client."""
    headers = {
        "Authorization": f"Bearer {os.getenv('CF_API_KEY')}",
        "Content-Type": "application/json",
    }
    client = aiohttp.ClientSession(headers=headers)
    client.CF_ACCOUNT_ENDPOINT = CF_ACCOUNT_ENDPOINT.format(
        CF_API_URL=CF_API_URL, CF_ACCOUNT_ID=os.getenv("CF_ACCOUNT_ID")
    )
    client.CF_TUNNEL_ENDPOINT = CF_TUNNEL_ENDPOINT.format(
        CF_ACCOUNT_ENDPOINT=client.CF_ACCOUNT_ENDPOINT,
        CF_TUNNEL_ID=os.getenv("CF_TUNNEL_ID"),
    )
    client.CF_DNS_ENDPOINT = CF_DNS_ENDPOINT.format(
        CF_API_URL=CF_API_URL, CF_ZONE_ID=os.getenv("CF_ZONE_ID")
    )
    client.CF_TUNNEL_DNS = f"{os.getenv('CF_TUNNEL_ID')}.cfargotunnel.com"
    return client


async def create_dns_record(hostname: str) -> None:
    """Create a DNS record."""
    async with get_client() as client:
        domain = os.getenv("CF_DOMAIN")
        if hostname.endswith("."):
            hostname = hostname[:-1]
        if not hostname.endswith(domain):
            hostname = f"{hostname}.{domain}"
        async with client.post(
            client.CF_DNS_ENDPOINT,
            json={
                "type": "CNAME",
                "name": hostname,
                "content": client.CF_TUNNEL_DNS,
                "ttl": 1,
                "proxied": True,
            },
        ) as resp:
            return


async def get_tunnel_config() -> list[dict]:
    """Get the tunnel configuration."""
    async with get_client() as client:
        async with client.get(f"{client.CF_TUNNEL_ENDPOINT}/configurations") as resp:
            tunnels = await resp.json()
            return tunnels["result"]["config"]["ingress"]


def add_tunnel_config(
    tunnels: list[dict], service: str, hostname: str, originRequest: dict = None
) -> None:
    """Add a tunnel configuration."""
    for tunnel in tunnels:
        if "hostname" in tunnel and hostname in tunnel["hostname"]:
            raise IndexError(f"Hostname {hostname} already exists.")
    else:
        domain = os.getenv("CF_DOMAIN")
        if hostname.endswith("."):
            hostname = hostname[:-1]
        if not hostname.endswith(domain):
            hostname = f"{hostname}.{domain}"
        ingress = {
            "service": service,
            "hostname": hostname,
            "originRequest": originRequest if originRequest else {},
        }
        tunnels.insert(0, ingress)


async def update_tunnel_config(tunnels: dict[Ingress]) -> None:
    """Update the tunnel configuration."""
    async with get_client() as client:
        async with client.put(
            f"{client.CF_TUNNEL_ENDPOINT}/configurations",
            json={"config": {"ingress": tunnels}},
        ) as resp:
            print(await resp.json())



# Collections of utility functions

import string
import secrets


def try_except(success, failure, *exceptions):
    """Try to call a function, if it fails, call another function or return a value."""
    for e in exceptions:
        print(e)
    try:
        return success()
    except exceptions or Exception:
        return failure() if callable(failure) else failure


def generate_password(length: int = 32):
    """Generate a random password."""
    chars = string.ascii_letters + string.digits + "-_!@#$%^&*()"
    return "".join(secrets.choice(chars) for i in range(length))



import os

import interactions

from dotenv import load_dotenv

from midgard_discord import cloud
from midgard_discord import commands
from midgard_discord import database
from midgard_discord import networking
from midgard_discord import texts


# Setup discord API
load_dotenv()

CACHE_TTL = os.getenv("DB_CACHE_TTL")
bot = interactions.Client(
    token=os.getenv("DISCORD_TOKEN"),
    default_scope=os.getenv("DISCORD_DEFAULT_GUILD_ID").split(","),
)


@bot.event
async def on_ready():
    """This event is called when the bot is ready to start accepting commands."""
    print(f"We're online! We've logged in as {bot.me.name}.")
    print(f"Our latency is {round(bot.latency)} ms.")


@bot.command(name="midgard")
async def midgard(ctx: interactions.CommandContext):
    """Midgard group command"""
    pass


@midgard.subcommand(name="help", description="Get help with Midgard")
async def help(ctx: interactions.CommandContext):
    """Get help with Midgard"""
    await commands.help(ctx)


@midgard.subcommand(name="register", description="Register an account in Midgard")
@interactions.autodefer(delay=2.5)
async def register(ctx: interactions.CommandContext):
    """Request enrolment to Midgard"""
    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    os_client = cloud.connect()

    await commands.register(ctx, db_session, os_client)

    # Close database connection
    await db_engine.dispose()
    os_client.close()


@midgard.group(name="add")
async def add(ctx: interactions.CommandContext):
    """Set group command"""
    pass


@add.subcommand(
    name="keypair",
    description="Add your SSH-public key in Midgard",
    options=[
        interactions.Option(
            name="public_key",
            description="Your SSH public key",
            type=interactions.OptionType.STRING,
            required=True,
        ),
    ],
)
async def keypair(ctx: interactions.CommandContext, public_key: str):
    """Set your SSH-public key in Midgard"""
    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    user = await database.find_user(db_session, str(ctx.author.user.id))

    if user is None:
        await ctx.send(
            texts.ERROR_NOT_REGISTERED.format(discord_user_id=ctx.author.user.id)
        )
    else:
        os_client = cloud.connect(
            auth_url=os.getenv("OS_AUTH_URL"),
            region_name=os.getenv("OS_REGION_NAME"),
            project_name=user.project_name,
            username=user.username,
            password=user.password,
            user_domain=os.getenv("OS_USER_DOMAIN_NAME"),
            project_domain=os.getenv("OS_PROJECT_DOMAIN_NAME"),
        )

        keypair = await cloud.find_keypair(os_client)
        # Create a new keypair if it doesn't exist
        if keypair is None:
            try:
                await cloud.create_keypair(os_client, public_key)
                await ctx.send(
                    f"<@{ctx.author.user.id}> Your public key has been updated!"
                )
            except Exception as e:
                await ctx.send(f"<@{ctx.author.user.id}> {e}")
        # Update the keypair if it exists
        else:
            await cloud.delete_keypair(os_client, keypair)
            try:
                await cloud.create_keypair(os_client, public_key)
                await ctx.send(
                    f"<@{ctx.author.user.id}> Your public key has been updated!"
                )
            except Exception as e:
                await ctx.send(f"<@{ctx.author.user.id}> {e}")
    # Close database connection
    await db_engine.dispose()
    os_client.close()


@add.subcommand(
    name="portforward",
    description="Add port forwarding rules in Midgard",
    options=[
        interactions.Option(
            name="port",
            description="The port you want to forward",
            type=interactions.OptionType.INTEGER,
            required=True,
        ),
        interactions.Option(
            name="protocol",
            description="The protocol you want to forward",
            type=interactions.OptionType.STRING,
            choices=[
                interactions.Choice(name="TCP", value="tcp"),
                interactions.Choice(name="HTTP", value="http"),
                interactions.Choice(name="SSH", value="ssh"),
            ],
            default="http",
        ),
    ],
)
async def portforward(
    ctx: interactions.CommandContext, port: int, protocol: str = "http"
):
    """Add port forwarding rules to security group in Midgard"""
    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    user = await database.find_user(db_session, str(ctx.author.user.id))

    if user is None:
        await ctx.send(
            texts.ERROR_NOT_REGISTERED.format(discord_user_id=ctx.author.user.id)
        )
    else:
        os_client = cloud.connect(
            auth_url=os.getenv("OS_AUTH_URL"),
            region_name=os.getenv("OS_REGION_NAME"),
            project_name=user.project_name,
            username=user.username,
            password=user.password,
            user_domain=os.getenv("OS_USER_DOMAIN_NAME"),
            project_domain=os.getenv("OS_PROJECT_DOMAIN_NAME"),
        )

        server = await cloud.find_server(os_client)
        if server is None:
            await ctx.send(
                texts.ERROR_SERVER_NOT_FOUND.format(discord_user_id=ctx.author.user.id)
            )
        else:
            try:
                tunnels = await networking.get_tunnel_config()
                await cloud.add_security_group_rule(os_client, port)
                public_ip = [
                    ip["addr"]
                    for ip in server.addresses["default"]
                    if ip["OS-EXT-IPS:type"] == "floating"
                ][0]
                service = f"{protocol}://{public_ip}:{port}"
                hostname = f"{user.username}-{protocol}-{port}"
                networking.add_tunnel_config(tunnels, service, hostname)
                await networking.update_tunnel_config(tunnels)
                await networking.create_dns_record(hostname)
                await ctx.send(
                    f"<@{ctx.author.user.id}> Port forwarding rule has been added!",
                    ephemeral=True,
                )
            except Exception as e:
                await ctx.send(f"<@{ctx.author.user.id}> {e}")

    # Close database connection
    await db_engine.dispose()
    os_client.close()


# @add.subcommand(
#     name="dns",
#     description="Add a DNS tunnel to your server",
#     options=[
#         interactions.Option(
#             name="hostname",
#             description="The hostname you want to forward",
#             type=interactions.OptionType.STRING,
#             required=True,
#         ),
#         interactions.Option(
#             name="port",
#             description="The port you want to forward",
#             type=interactions.OptionType.INTEGER,
#             required=True,
#         ),
#         interactions.Option(
#             name="protocol",
#             description="The protocol you want to forward",
#             type=interactions.OptionType.STRING,
#             choices=[
#                 interactions.Choice(name="TCP", value="tcp"),
#                 interactions.Choice(name="HTTP", value="http"),
#                 interactions.Choice(name="SSH", value="ssh"),
#             ],
#             default="http",
#         ),
#     ],
# )
async def add_dns(
    ctx: interactions.CommandContext, hostname: str, port: int, protocol: str = "http"
):
    """Add a DNS tunnel to your server"""

    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    user = await database.find_user(db_session, str(ctx.author.user.id))

    if user is None:
        await ctx.send(
            texts.ERROR_NOT_REGISTERED.format(discord_user_id=ctx.author.user.id)
        )
    else:
        tunnels = await networking.get_tunnel_config()

        try:
            os_client = cloud.connect(
                auth_url=os.getenv("OS_AUTH_URL"),
                region_name=os.getenv("OS_REGION_NAME"),
                project_name=user.project_name,
                username=user.username,
                password=user.password,
                user_domain=os.getenv("OS_USER_DOMAIN_NAME"),
                project_domain=os.getenv("OS_PROJECT_DOMAIN_NAME"),
            )

            server = await cloud.find_server(os_client)
            public_ip = [
                ip["addr"]
                for ip in server.addresses["default"]
                if ip["OS-EXT-IPS:type"] == "floating"
            ][0]
            service = f"{protocol}://{public_ip}:{port}"
            networking.add_tunnel_config(tunnels, service, hostname)
            await networking.update_tunnel_config(tunnels)
            await networking.create_dns_record(hostname)
            await ctx.send(
                f"<@{ctx.author.user.id}> DNS tunnel has been added to your server!",
                ephemeral=True,
            )
        except Exception as e:
            await ctx.send(f"<@{ctx.author.user.id}> {e}")

    # Close database connection
    await db_engine.dispose()
    os_client.close()


@midgard.group(name="server")
async def server(ctx: interactions.CommandContext):
    """Server group command"""
    pass


@server.subcommand(
    name="launch",
    description="Create a VM server",
    options=[
        interactions.Option(
            name="flavor",
            description="The flavor of the server",
            type=interactions.OptionType.STRING,
            required=True,
            autocomplete=True,
        ),
        interactions.Option(
            name="image",
            description="The image of the server",
            type=interactions.OptionType.STRING,
            required=True,
            autocomplete=True,
        ),
    ],
)
@interactions.autodefer()
async def server_launch(ctx: interactions.CommandContext, flavor: str, image: str):
    """Create a VM server"""
    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    user = await database.find_user(db_session, str(ctx.author.user.id))

    if user is None:
        await ctx.send(
            texts.ERROR_NOT_REGISTERED.format(discord_user_id=ctx.author.user.id)
        )
    else:
        os_client = cloud.connect(
            auth_url=os.getenv("OS_AUTH_URL"),
            region_name=os.getenv("OS_REGION_NAME"),
            project_name=user.project_name,
            username=user.username,
            password=user.password,
            user_domain=os.getenv("OS_USER_DOMAIN_NAME"),
            project_domain=os.getenv("OS_PROJECT_DOMAIN_NAME"),
        )

        server = await cloud.find_server(os_client)
        if server is not None:
            if server.status == "ACTIVE":
                await ctx.send(
                    f"<@{ctx.author.user.id}> Your server is already running."
                )
            else:
                await ctx.send(
                    f"<@{ctx.author.user.id}> Your server is in STATUS {server.status}."
                )
        else:
            keypair = await cloud.find_keypair(os_client)
            security_group = await cloud.find_default_security_group(os_client)
            try:
                # Check if keypair exists
                if keypair is None:
                    raise Exception(
                        texts.ERROR_KEYPAIR_NOT_FOUND.format(
                            discord_user_id=ctx.author.user.id
                        )
                    )
                # Create a Nova server
                server = await cloud.create_server(
                    os_client,
                    key_name=keypair.name,
                    flavor=flavor,
                    image=image,
                    auto_ip=True,
                    security_groups=[security_group.name],
                    reuse_ips=True,
                    wait=True,
                )
                # Set up SSH tunnel
                tunnels = await networking.get_tunnel_config()
                await cloud.add_security_group_rule(os_client, 22)
                public_ip = [
                    ip["addr"]
                    for ip in server.addresses["default"]
                    if ip["OS-EXT-IPS:type"] == "floating"
                ][0]
                service = f"ssh://{public_ip}:22"
                hostname = f"{user.username}-ssh.{os.getenv('CF_DOMAIN')}"
                networking.add_tunnel_config(tunnels, service, hostname)
                await networking.update_tunnel_config(tunnels)
                await networking.create_dns_record(hostname)
                await ctx.send(
                    texts.SERVER_CREATED.format(
                        discord_user_id=ctx.author.user.id,
                        server_name=server.name,
                        server_id=server.id,
                        hostname=hostname,
                    ),
                    ephemeral=True,
                )
            except Exception as e:
                await ctx.send(f"<@{ctx.author.user.id}> {e}")

    # Close database connection
    await db_engine.dispose()
    os_client.close()


@server_launch.autocomplete("flavor")
async def server_launch_flavor_autocomplete(
    ctx: interactions.CommandContext, user_input: str = ""
):
    """Autocomplete for create server flavor"""
    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    user = await database.find_user(db_session, str(ctx.author.user.id))

    if user is None:
        await ctx.send(
            texts.ERROR_NOT_REGISTERED.format(discord_user_id=ctx.author.user.id)
        )
        return

    os_client = cloud.connect(
        auth_url=os.getenv("OS_AUTH_URL"),
        region_name=os.getenv("OS_REGION_NAME"),
        project_name=user.project_name,
        username=user.username,
        password=user.password,
        user_domain=os.getenv("OS_USER_DOMAIN_NAME"),
        project_domain=os.getenv("OS_PROJECT_DOMAIN_NAME"),
    )

    flavors = await cloud.list_flavors(os_client)

    choices = [
        interactions.Choice(
            name=f"{flavor.name} ({flavor.vcpus} vCPUs, {flavor.ram/1024}GB RAM, {flavor.disk}GB HDD)",
            value=flavor.id,
        )
        for flavor in flavors
        if user_input in flavor.name and flavor.vcpus <= 2
    ]
    # Close database connection
    await db_engine.dispose()
    os_client.close()
    await ctx.populate(choices)


@server_launch.autocomplete("image")
async def server_launch_image_autocomplete(
    ctx: interactions.CommandContext, user_input: str = ""
):
    """Autocomplete for create server image"""
    db_engine, db_session = await database.init_async_db(os.getenv("DB_URI"))
    user = await database.find_user(db_session, str(ctx.author.user.id))

    os_client = cloud.connect(
        auth_url=os.getenv("OS_AUTH_URL"),
        region_name=os.getenv("OS_REGION_NAME"),
        project_name=user.project_name,
        username=user.username,
        password=user.password,
        user_domain=os.getenv("OS_USER_DOMAIN_NAME"),
        project_domain=os.getenv("OS_PROJECT_DOMAIN_NAME"),
    )

    images = await cloud.list_images(os_client)

    choices = [
        interactions.Choice(name=image.name, value=image.id)
        for image in images
        if user_input in image.name
    ]
    # Close database connection
    await db_engine.dispose()
    os_client.close()
    await ctx.populate(choices)


def main():
    """Main function"""
    bot.start()


if __name__ == "__main__":
    """Run main function"""
    main()



# Large texts in the application


# Fragments
INFO_MORE = """
For more information, visit https://docs.midgardlab.io
"""


# Success texts
WELCOME = (
    """
Welcome to Midgard. Midgard is a private OpenStack cloud for VAIT testing, training, and development purpose.
At the moment, Midgard offers the following capacities:
    1. Virtual Machines.
    2. Automatic SSL certificate termination.
"""
    + INFO_MORE
)

REGISTERED = (
    """
<@{discord_user_id}> You have been successfully registered.
Add an SSH public key to your account by running `/midgard keypair add`. This is required to access your server.
You can now create a server by running `/midgard server launch`.
"""
    + INFO_MORE
)

PORT_FORWARDED = (
    """
<@{discord_user_id}> `{protocol}://{server_ip}:{port}` successfully deployed.`. Access it at:

https://{hostname}.midgardlab.io
"""
    + INFO_MORE
)

SERVER_CREATED = (
    """
<@{discord_user_id}> Your server has been successfully created. To access your server, add the following to your `~/.ssh/config` file:
```
Host {server_name}
    HostName {server_ip}
    ProxyCommand /usr/local/bin/cloudflared access ssh --hostname {hostname}
```
Then you can access your server by running `ssh {server_name}`.
"""
    + INFO_MORE
)


# Error texts
ERROR_REGISTERED = (
    """
<@{discord_user_id}> You are already registered.
Add an SSH public key to your account by running `/midgard keypair add`
Or run `/midgard server launch` to create a server.
"""
    + INFO_MORE
)

ERROR_NOT_REGISTERED = (
    """
<@{discord_user_id}> You are not yet registered. Please register by running `/migard register`.
"""
    + INFO_MORE
)

ERROR_SERVER_NOT_FOUND = (
    """
<@{discord_user_id}> You do not have any server. Please create a server by running `/midgard server launch`.
"""
    + INFO_MORE
)

ERROR_KEYPAIR_NOT_FOUND = (
    """
<@{discord_user_id}> You do not have any SSH keypair. Please add an SSH keypair by running `/midgard add keypair`.
"""
    + INFO_MORE
)



# Cache database functions

from sqlalchemy import select
from sqlalchemy import DateTime
from sqlalchemy.sql import func
from sqlalchemy.orm import DeclarativeBase, Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker


class Base(DeclarativeBase):
    """Base class for all models."""

    pass


class OpenStackCredential(Base):
    """OpenStack RC Credential model."""

    __tablename__ = "openstack_credentials"

    username: Mapped[str] = mapped_column(primary_key=True)
    password: Mapped[str]
    project_name: Mapped[str]
    created_at: Mapped[DateTime] = mapped_column(DateTime(), server_default=func.now())
    updated_at: Mapped[DateTime] = mapped_column(DateTime(), server_default=func.now())

    def __repr__(self):
        """Return a string representation of the model."""
        return f"<OpenStackCredential(username={self.username}, project_name={self.project_name}>"

    def __getitem__(self, field):
        """Return the value of a field."""
        return self.__dict__[field]


async def init_async_db(DB_URI: str) -> tuple[create_async_engine, async_sessionmaker]:
    """Initialise the database."""
    engine = create_async_engine(DB_URI)

    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    session = async_sessionmaker(engine, expire_on_commit=False)
    return engine, session


async def find_user(
    async_session: async_sessionmaker[AsyncSession], discord_user_id: str
):
    """Find a user the Database"""
    stmt = select(OpenStackCredential).where(
        OpenStackCredential.username == discord_user_id
    )
    async with async_session() as session:
        result = (await session.execute(stmt)).first()
        return result[0] if result is not None else None


async def create_user(
    async_session: async_sessionmaker[AsyncSession], discord_user_id: str, **kawrgs
) -> None:
    """Create a new user in the Database"""
    user = OpenStackCredential(username=discord_user_id, **kawrgs)
    async with async_session() as session:
        async with session.begin():
            session.add(user)
    await session.commit()



# Internal commands module for Midgard Discord Bot
import os
import interactions
import openstack
import sqlalchemy

from datetime import datetime

from midgard_discord import cloud
from midgard_discord import database
from midgard_discord import networking
from midgard_discord import texts
from midgard_discord import utils


def log(who: str, event: str) -> None:
    now = datetime.now()
    timestamp = now.strftime("%Y-%m-%d %H:%M:%S")
    print(f"[{timestamp}] <{who}> {event}")


async def help(ctx: interactions.CommandContext):
    """Send a welcome message."""
    log(ctx.author.name, "/midgard help")
    await ctx.send(texts.WELCOME)


async def register(
    ctx: interactions.CommandContext,
    db_session: sqlalchemy.ext.asyncio.async_sessionmaker,
    os_client: openstack.connection.Connection,
):
    """Register a user to Midgard."""
    log(ctx.author.name, "/midgard register")
    user = await database.find_user(db_session, str(ctx.author.user.id))

    project_name = f"{os.getenv('OS_DEFAULT_GUILD_PREFIX')}_{ctx.author.user.id}"
    # If we miss the cache, check the database
    if user is None:
        os_user = await cloud.find_user(os_client, str(ctx.author.user.id))
        # If we miss the database, create the user
        if os_user is None:
            # Create user and project in OpenStack
            os_project = await cloud.create_project(os_client, project_name)
            user_password = utils.generate_password()
            os_user = await cloud.create_user(
                os_client,
                str(ctx.author.user.id),
                default_project=os_project,
                password=user_password,
            )

            # # Set roles
            await cloud.set_default_roles(os_client, os_user, os_project)

            # Setup default network
            await cloud.setup_default_network(os_client, os_project)

            # Cache user in database
            await database.create_user(
                db_session,
                str(ctx.author.user.id),
                password=user_password,
                project_name=project_name,
            )
        # If we find the user in OpenStack, reset the password and cache it in database
        else:
            user_password = utils.generate_password()
            await cloud.update_user(os_client, os_user, password=user_password)
            project = await cloud.find_project(os_client, project_name)
            await database.create_user(
                db_session,
                str(ctx.author.user.id),
                password=user_password,
                project_name=project.name,
            )
        await ctx.send(texts.REGISTERED.format(discord_user_id=ctx.author.user.id))
    else:
        await ctx.send(
            texts.ERROR_REGISTERED.format(discord_user_id=ctx.author.user.id)
        )
